@@ -1,5 +1,5 @@
 /*
-    Copyright (c) 2005-2021 Intel Corporation
+    Copyright (c) 2005-2025 Intel Corporation
 
     Licensed under the Apache License, Version 2.0 (the "License");
     you may not use this file except in compliance with the License.
@@ -16,6 +16,7 @@
 
 #include "task_dispatcher.h"
 #include "governor.h"
+#include "threading_control.h"
 #include "arena.h"
 #include "itt_notify.h"
 #include "semaphore.h"
@@ -59,25 +60,101 @@ numa_binding_observer* construct_binding_observer( d1::task_arena* ta, int num_s
     if ((core_type >= 0 && core_type_count() > 1) || (numa_id >= 0 && numa_node_count() > 1) || max_threads_per_core > 0) {
         binding_observer = new(allocate_memory(sizeof(numa_binding_observer))) numa_binding_observer(ta, num_slots, numa_id, core_type, max_threads_per_core);
         __TBB_ASSERT(binding_observer, "Failure during NUMA binding observer allocation and construction");
-        binding_observer->observe(true);
     }
     return binding_observer;
 }
 
 void destroy_binding_observer( numa_binding_observer* binding_observer ) {
-    __TBB_ASSERT(binding_observer, "Trying to deallocate NULL pointer");
+    __TBB_ASSERT(binding_observer, "Trying to deallocate nullptr pointer");
     binding_observer->observe(false);
     binding_observer->~numa_binding_observer();
     deallocate_memory(binding_observer);
 }
 #endif /*!__TBB_ARENA_BINDING*/
 
+void arena::on_thread_leaving(unsigned ref_param) {
+    //
+    // Implementation of arena destruction synchronization logic contained various
+    // bugs/flaws at the different stages of its evolution, so below is a detailed
+    // description of the issues taken into consideration in the framework of the
+    // current design.
+    //
+    // In case of using fire-and-forget tasks (scheduled via task::enqueue())
+    // external thread is allowed to leave its arena before all its work is executed,
+    // and market may temporarily revoke all workers from this arena. Since revoked
+    // workers never attempt to reset arena state to EMPTY and cancel its request
+    // to RML for threads, the arena object is destroyed only when both the last
+    // thread is leaving it and arena's state is EMPTY (that is its external thread
+    // left and it does not contain any work).
+    // Thus resetting arena to EMPTY state (as earlier TBB versions did) should not
+    // be done here (or anywhere else in the external thread to that matter); doing so
+    // can result either in arena's premature destruction (at least without
+    // additional costly checks in workers) or in unnecessary arena state changes
+    // (and ensuing workers migration).
+    //
+    // A worker that checks for work presence and transitions arena to the EMPTY
+    // state (in snapshot taking procedure arena::out_of_work()) updates
+    // arena::my_pool_state first and only then arena::my_num_workers_requested.
+    // So the check for work absence must be done against the latter field.
+    //
+    // In a time window between decrementing the active threads count and checking
+    // if there is an outstanding request for workers. New worker thread may arrive,
+    // finish remaining work, set arena state to empty, and leave decrementing its
+    // refcount and destroying. Then the current thread will destroy the arena
+    // the second time. To preclude it a local copy of the outstanding request
+    // value can be stored before decrementing active threads count.
+    //
+    // But this technique may cause two other problem. When the stored request is
+    // zero, it is possible that arena still has threads and they can generate new
+    // tasks and thus re-establish non-zero requests. Then all the threads can be
+    // revoked (as described above) leaving this thread the last one, and causing
+    // it to destroy non-empty arena.
+    //
+    // The other problem takes place when the stored request is non-zero. Another
+    // thread may complete the work, set arena state to empty, and leave without
+    // arena destruction before this thread decrements the refcount. This thread
+    // cannot destroy the arena either. Thus the arena may be "orphaned".
+    //
+    // In both cases we cannot dereference arena pointer after the refcount is
+    // decremented, as our arena may already be destroyed.
+    //
+    // If this is the external thread, the market is protected by refcount to it.
+    // In case of workers market's liveness is ensured by the RML connection
+    // rundown protocol, according to which the client (i.e. the market) lives
+    // until RML server notifies it about connection termination, and this
+    // notification is fired only after all workers return into RML.
+    //
+    // Thus if we decremented refcount to zero we ask the market to check arena
+    // state (including the fact if it is alive) under the lock.
+    //
+
+    __TBB_ASSERT(my_references.load(std::memory_order_relaxed) >= ref_param, "broken arena reference counter");
+
+    // When there is no workers someone must free arena, as
+    // without workers, no one calls out_of_work().
+    if (ref_param == ref_external && !my_mandatory_concurrency.test()) {
+        out_of_work();
+    }
+
+    threading_control* tc = my_threading_control;
+    auto tc_client_snapshot = tc->prepare_client_destruction(my_tc_client);
+    // Release our reference to sync with destroy_client
+    unsigned remaining_ref = my_references.fetch_sub(ref_param, std::memory_order_release) - ref_param;
+    // do not access `this` it might be destroyed already
+    if (remaining_ref == 0) {
+        if (tc->try_destroy_client(tc_client_snapshot)) {
+            // We are requested to destroy ourself
+            free_arena();
+        }
+    }
+}
+
 std::size_t arena::occupy_free_slot_in_range( thread_data& tls, std::size_t lower, std::size_t upper ) {
     if ( lower >= upper ) return out_of_arena;
     // Start search for an empty slot from the one we occupied the last time
     std::size_t index = tls.my_arena_index;
     if ( index < lower || index >= upper ) index = tls.my_random.get() % (upper - lower) + lower;
-    __TBB_ASSERT( index >= lower && index < upper, NULL );
+    __TBB_ASSERT( index >= lower && index < upper, nullptr);
     // Find a free slot
     for ( std::size_t i = index; i < upper; ++i )
         if (my_slots[i].try_occupy()) return i;
@@ -104,19 +181,20 @@ std::size_t arena::occupy_free_slot(thread_data& tls) {
 
 std::uintptr_t arena::calculate_stealing_threshold() {
     stack_anchor_type anchor;
-    return r1::calculate_stealing_threshold(reinterpret_cast<std::uintptr_t>(&anchor), my_market->worker_stack_size());
+    return r1::calculate_stealing_threshold(reinterpret_cast<std::uintptr_t>(&anchor), my_threading_control->worker_stack_size());
 }
 
 void arena::process(thread_data& tls) {
     governor::set_thread_data(tls); // TODO: consider moving to create_one_job.
     __TBB_ASSERT( is_alive(my_guard), nullptr);
-    __TBB_ASSERT( my_num_slots > 1, nullptr);
+    __TBB_ASSERT( my_num_slots >= 1, nullptr);
 
     std::size_t index = occupy_free_slot</*as_worker*/true>(tls);
     if (index == out_of_arena) {
-        on_thread_leaving<ref_worker>();
+        on_thread_leaving(ref_worker);
         return;
     }
+
     __TBB_ASSERT( index >= my_num_reserved_slots, "Workers cannot occupy reserved slots" );
     tls.attach_arena(*this, index);
     // worker thread enters the dispatch loop to look for a work
@@ -126,9 +204,8 @@ void arena::process(thread_data& tls) {
     }
 
     task_dispatcher& task_disp = tls.my_arena_slot->default_task_dispatcher();
-    task_disp.set_stealing_threshold(calculate_stealing_threshold());
+    tls.enter_task_dispatcher(task_disp, calculate_stealing_threshold());
     __TBB_ASSERT(task_disp.can_steal(), nullptr);
-    tls.attach_task_dispatcher(task_disp);
 
     __TBB_ASSERT( !tls.my_last_observer, "There cannot be notified local observers when entering arena" );
     my_observers.notify_entry_observers(tls.my_last_observer, tls.my_is_worker);
@@ -147,8 +224,7 @@ void arena::process(thread_data& tls) {
     my_observers.notify_exit_observers(tls.my_last_observer, tls.my_is_worker);
     tls.my_last_observer = nullptr;
 
-    task_disp.set_stealing_threshold(0);
-    tls.detach_task_dispatcher();
+    tls.leave_task_dispatcher();
 
     // Arena slot detach (arena may be used in market::process)
     // TODO: Consider moving several calls below into a new method(e.g.detach_arena).
@@ -161,36 +237,39 @@ void arena::process(thread_data& tls) {
     // In contrast to earlier versions of TBB (before 3.0 U5) now it is possible
     // that arena may be temporarily left unpopulated by threads. See comments in
     // arena::on_thread_leaving() for more details.
-    on_thread_leaving<ref_worker>();
+    on_thread_leaving(ref_worker);
     __TBB_ASSERT(tls.my_arena == this, "my_arena is used as a hint when searching the arena to join");
 }
 
-arena::arena ( market& m, unsigned num_slots, unsigned num_reserved_slots, unsigned priority_level )
+arena::arena(threading_control* control, unsigned num_slots, unsigned num_reserved_slots, unsigned priority_level
+#if __TBB_PREVIEW_PARALLEL_PHASE
+             , tbb::task_arena::leave_policy lp 
+#endif
+)
 {
     __TBB_ASSERT( !my_guard, "improperly allocated arena?" );
     __TBB_ASSERT( sizeof(my_slots[0]) % cache_line_size()==0, "arena::slot size not multiple of cache line size" );
     __TBB_ASSERT( is_aligned(this, cache_line_size()), "arena misaligned" );
-    my_market = &m;
+    my_threading_control = control;
     my_limit = 1;
     // Two slots are mandatory: for the external thread, and for 1 worker (required to support starvation resistant tasks).
-    my_num_slots = num_arena_slots(num_slots);
+    my_num_slots = num_arena_slots(num_slots, num_reserved_slots);
     my_num_reserved_slots = num_reserved_slots;
     my_max_num_workers = num_slots-num_reserved_slots;
     my_priority_level = priority_level;
     my_references = ref_external; // accounts for the external thread
-    my_aba_epoch = m.my_arenas_aba_epoch.load(std::memory_order_relaxed);
     my_observers.my_arena = this;
     my_co_cache.init(4 * num_slots);
-    __TBB_ASSERT ( my_max_num_workers <= my_num_slots, NULL );
+    __TBB_ASSERT ( my_max_num_workers <= my_num_slots, nullptr);
     // Initialize the default context. It should be allocated before task_dispatch construction.
     my_default_ctx = new (cache_aligned_allocate(sizeof(d1::task_group_context)))
         d1::task_group_context{ d1::task_group_context::isolated, d1::task_group_context::fp_settings };
     // Construct slots. Mark internal synchronization elements for the tools.
     task_dispatcher* base_td_pointer = reinterpret_cast<task_dispatcher*>(my_slots + my_num_slots);
     for( unsigned i = 0; i < my_num_slots; ++i ) {
-        // __TBB_ASSERT( !my_slots[i].my_scheduler && !my_slots[i].task_pool, NULL );
-        __TBB_ASSERT( !my_slots[i].task_pool_ptr, NULL );
-        __TBB_ASSERT( !my_slots[i].my_task_pool_size, NULL );
+        // __TBB_ASSERT( !my_slots[i].my_scheduler && !my_slots[i].task_pool, nullptr);
+        __TBB_ASSERT( !my_slots[i].task_pool_ptr, nullptr);
+        __TBB_ASSERT( !my_slots[i].my_task_pool_size, nullptr);
         mailbox(i).construct();
         my_slots[i].init_task_streams(i);
         my_slots[i].my_default_task_dispatcher = new(base_td_pointer + i) task_dispatcher(this);
@@ -198,39 +277,44 @@ arena::arena ( market& m, unsigned num_slots, unsigned num_reserved_slots, unsig
     }
     my_fifo_task_stream.initialize(my_num_slots);
     my_resume_task_stream.initialize(my_num_slots);
-#if __TBB_PREVIEW_CRITICAL_TASKS
+#if __TBB_CRITICAL_TASKS
     my_critical_task_stream.initialize(my_num_slots);
 #endif
-#if __TBB_ENQUEUE_ENFORCED_CONCURRENCY
-    my_local_concurrency_requests = 0;
-    my_local_concurrency_flag.clear();
-    my_global_concurrency_mode.store(false, std::memory_order_relaxed);
+    my_mandatory_requests = 0;
+
+#if __TBB_PREVIEW_PARALLEL_PHASE
+    my_thread_leave.set_initial_state(lp);
 #endif
 }
 
-arena& arena::allocate_arena( market& m, unsigned num_slots, unsigned num_reserved_slots,
-                              unsigned priority_level )
+arena& arena::allocate_arena(threading_control* control, unsigned num_slots, unsigned num_reserved_slots,
+                             unsigned priority_level
+#if __TBB_PREVIEW_PARALLEL_PHASE
+                             , tbb::task_arena::leave_policy lp
+#endif
+)
 {
     __TBB_ASSERT( sizeof(base_type) + sizeof(arena_slot) == sizeof(arena), "All arena data fields must go to arena_base" );
     __TBB_ASSERT( sizeof(base_type) % cache_line_size() == 0, "arena slots area misaligned: wrong padding" );
     __TBB_ASSERT( sizeof(mail_outbox) == max_nfs_size, "Mailbox padding is wrong" );
-    std::size_t n = allocation_size(num_arena_slots(num_slots));
+    std::size_t n = allocation_size(num_arena_slots(num_slots, num_reserved_slots));
     unsigned char* storage = (unsigned char*)cache_aligned_allocate(n);
     // Zero all slots to indicate that they are empty
     std::memset( storage, 0, n );
-    return *new( storage + num_arena_slots(num_slots) * sizeof(mail_outbox) )
-        arena(m, num_slots, num_reserved_slots, priority_level);
+
+    return *new( storage + num_arena_slots(num_slots, num_reserved_slots) * sizeof(mail_outbox) )
+        arena(control, num_slots, num_reserved_slots, priority_level
+#if __TBB_PREVIEW_PARALLEL_PHASE
+              , lp
+#endif
+        );
 }
 
 void arena::free_arena () {
-    __TBB_ASSERT( is_alive(my_guard), NULL );
+    __TBB_ASSERT( is_alive(my_guard), nullptr);
     __TBB_ASSERT( !my_references.load(std::memory_order_relaxed), "There are threads in the dying arena" );
-    __TBB_ASSERT( !my_num_workers_requested && !my_num_workers_allotted, "Dying arena requests workers" );
-    __TBB_ASSERT( my_pool_state.load(std::memory_order_relaxed) == SNAPSHOT_EMPTY || !my_max_num_workers,
-                  "Inconsistent state of a dying arena" );
-#if __TBB_ENQUEUE_ENFORCED_CONCURRENCY
-    __TBB_ASSERT( !my_global_concurrency_mode, NULL );
-#endif
+    __TBB_ASSERT( !my_total_num_workers_requested && !my_num_workers_allotted, "Dying arena requests workers" );
+    __TBB_ASSERT( is_empty(), "Inconsistent state of a dying arena" );
 #if __TBB_ARENA_BINDING
     if (my_numa_binding_observer != nullptr) {
         destroy_binding_observer(my_numa_binding_observer);
@@ -241,8 +325,8 @@ void arena::free_arena () {
     for ( unsigned i = 0; i < my_num_slots; ++i ) {
         // __TBB_ASSERT( !my_slots[i].my_scheduler, "arena slot is not empty" );
         // TODO: understand the assertion and modify
-        // __TBB_ASSERT( my_slots[i].task_pool == EmptyTaskPool, NULL );
-        __TBB_ASSERT( my_slots[i].head == my_slots[i].tail, NULL ); // TODO: replace by is_quiescent_local_task_pool_empty
+        // __TBB_ASSERT( my_slots[i].task_pool == EmptyTaskPool, nullptr);
+        __TBB_ASSERT( my_slots[i].head == my_slots[i].tail, nullptr); // TODO: replace by is_quiescent_local_task_pool_empty
         my_slots[i].free_task_pool();
         mailbox(i).drain();
         my_slots[i].my_default_task_dispatcher->~task_dispatcher();
@@ -253,18 +337,14 @@ void arena::free_arena () {
     my_co_cache.cleanup();
     my_default_ctx->~task_group_context();
     cache_aligned_deallocate(my_default_ctx);
-#if __TBB_PREVIEW_CRITICAL_TASKS
+#if __TBB_CRITICAL_TASKS
     __TBB_ASSERT( my_critical_task_stream.empty(), "Not all critical tasks were executed");
 #endif
-    // remove an internal reference
-    my_market->release( /*is_public=*/false, /*blocking_terminate=*/false );
-
     // Clear enfources synchronization with observe(false)
     my_observers.clear();
 
     void* storage  = &mailbox(my_num_slots-1);
-    __TBB_ASSERT( my_references.load(std::memory_order_relaxed) == 0, NULL );
-    __TBB_ASSERT( my_pool_state.load(std::memory_order_relaxed) == SNAPSHOT_EMPTY || !my_max_num_workers, NULL );
+    __TBB_ASSERT( my_references.load(std::memory_order_relaxed) == 0, nullptr);
     this->~arena();
 #if TBB_USE_ASSERT > 1
     std::memset( storage, 0, allocation_size(my_num_slots) );
@@ -276,78 +356,100 @@ bool arena::has_enqueued_tasks() {
     return !my_fifo_task_stream.empty();
 }
 
-bool arena::is_out_of_work() {
-#if __TBB_ENQUEUE_ENFORCED_CONCURRENCY
-    if (my_local_concurrency_flag.try_clear_if([this] {
-        return !has_enqueued_tasks();
-    })) {
-        my_market->adjust_demand(*this, /* delta = */ -1, /* mandatory = */ true);
-    }
-#endif
+void arena::request_workers(int mandatory_delta, int workers_delta, bool wakeup_threads) {
+    my_threading_control->adjust_demand(my_tc_client, mandatory_delta, workers_delta);
 
+    if (wakeup_threads) {
+        // Notify all sleeping threads that work has appeared in the arena.
+        get_waiting_threads_monitor().notify([&] (market_context context) {
+            return this == context.my_arena_addr;
+        });
+    }
+}
+
+bool arena::has_tasks() {
     // TODO: rework it to return at least a hint about where a task was found; better if the task itself.
-    switch (my_pool_state.load(std::memory_order_acquire)) {
-    case SNAPSHOT_EMPTY:
-        return true;
-    case SNAPSHOT_FULL: {
-        // Use unique id for "busy" in order to avoid ABA problems.
-        const pool_state_t busy = pool_state_t(&busy);
-        // Helper for CAS execution
-        pool_state_t expected_state;
-
-        // Request permission to take snapshot
-        expected_state = SNAPSHOT_FULL;
-        if (my_pool_state.compare_exchange_strong(expected_state, busy)) {
-            // Got permission. Take the snapshot.
-            // NOTE: This is not a lock, as the state can be set to FULL at
-            //       any moment by a thread that spawns/enqueues new task.
-            std::size_t n = my_limit.load(std::memory_order_acquire);
-            // Make local copies of volatile parameters. Their change during
-            // snapshot taking procedure invalidates the attempt, and returns
-            // this thread into the dispatch loop.
-            std::size_t k;
-            for (k = 0; k < n; ++k) {
-                if (my_slots[k].task_pool.load(std::memory_order_relaxed) != EmptyTaskPool &&
-                    my_slots[k].head.load(std::memory_order_relaxed) < my_slots[k].tail.load(std::memory_order_relaxed))
-                {
-                    // k-th primary task pool is nonempty and does contain tasks.
-                    break;
-                }
-                if (my_pool_state.load(std::memory_order_acquire) != busy)
-                    return false; // the work was published
-            }
-            bool work_absent = k == n;
-            // Test and test-and-set.
-            if (my_pool_state.load(std::memory_order_acquire) == busy) {
-                bool no_stream_tasks = !has_enqueued_tasks() && my_resume_task_stream.empty();
-#if __TBB_PREVIEW_CRITICAL_TASKS
-                no_stream_tasks = no_stream_tasks && my_critical_task_stream.empty();
+    std::size_t n = my_limit.load(std::memory_order_acquire);
+    bool tasks_are_available = false;
+    for (std::size_t k = 0; k < n && !tasks_are_available; ++k) {
+        tasks_are_available = !my_slots[k].is_empty();
+    }
+    tasks_are_available = tasks_are_available || has_enqueued_tasks() || !my_resume_task_stream.empty();
+#if __TBB_CRITICAL_TASKS
+    tasks_are_available = tasks_are_available || !my_critical_task_stream.empty();
 #endif
-                work_absent = work_absent && no_stream_tasks;
-                if (work_absent) {
-                    // save current demand value before setting SNAPSHOT_EMPTY,
-                    // to avoid race with advertise_new_work.
-                    int current_demand = (int)my_max_num_workers;
-                    expected_state = busy;
-                    if (my_pool_state.compare_exchange_strong(expected_state, SNAPSHOT_EMPTY)) {
-                        // This thread transitioned pool to empty state, and thus is
-                        // responsible for telling the market that there is no work to do.
-                        my_market->adjust_demand(*this, -current_demand, /* mandatory = */ false);
-                        return true;
-                    }
-                    return false;
-                }
-                // Undo previous transition SNAPSHOT_FULL-->busy, unless another thread undid it.
-                expected_state = busy;
-                my_pool_state.compare_exchange_strong(expected_state, SNAPSHOT_FULL);
-            }
+    return tasks_are_available;
+}
+
+void arena::out_of_work() {
+    // We should try unset my_pool_state first due to keep arena invariants in consistent state
+    // Otherwise, we might have my_pool_state = false and my_mandatory_concurrency = true that is broken invariant
+    bool disable_mandatory = my_mandatory_concurrency.try_clear_if([this] { return !has_enqueued_tasks(); });
+    bool release_workers = my_pool_state.try_clear_if([this] { return !has_tasks(); });
+
+    if (disable_mandatory || release_workers) {
+        int mandatory_delta = disable_mandatory ? -1 : 0;
+        int workers_delta = release_workers ? -(int)my_max_num_workers : 0;
+
+        if (disable_mandatory && is_arena_workerless()) {
+            // We had set workers_delta to 1 when enabled mandatory concurrency, so revert it now
+            workers_delta = -1;
         }
-        return false;
+        request_workers(mandatory_delta, workers_delta);
     }
-    default:
-        // Another thread is taking a snapshot.
-        return false;
+}
+
+void arena::set_top_priority(bool is_top_priority) {
+    my_is_top_priority.store(is_top_priority, std::memory_order_relaxed);
+}
+
+bool arena::is_top_priority() const {
+    return my_is_top_priority.load(std::memory_order_relaxed);
+}
+
+bool arena::try_join() {
+    if (is_joinable()) {
+        my_references += arena::ref_worker;
+        return true;
     }
+    return false;
+}
+
+void arena::set_allotment(unsigned allotment) {
+    if (my_num_workers_allotted.load(std::memory_order_relaxed) != allotment) {
+        my_num_workers_allotted.store(allotment, std::memory_order_relaxed);
+    }
+}
+
+int arena::update_concurrency(unsigned allotment) {
+    int delta = allotment - my_num_workers_allotted.load(std::memory_order_relaxed);
+    if (delta != 0) {
+        my_num_workers_allotted.store(allotment, std::memory_order_relaxed);
+    }
+    return delta;
+}
+
+std::pair<int, int> arena::update_request(int mandatory_delta, int workers_delta) {
+    __TBB_ASSERT(-1 <= mandatory_delta && mandatory_delta <= 1, nullptr);
+
+    int min_workers_request = 0;
+    int max_workers_request = 0;
+
+    // Calculate min request
+    my_mandatory_requests += mandatory_delta;
+    min_workers_request = my_mandatory_requests > 0 ? 1 : 0;
+
+    // Calculate max request
+    my_total_num_workers_requested += workers_delta;
+    // Clamp worker request into interval [0, my_max_num_workers]
+    max_workers_request = clamp(my_total_num_workers_requested, 0,
+        min_workers_request > 0 && is_arena_workerless() ? 1 : (int)my_max_num_workers);
+
+    return { min_workers_request, max_workers_request };
+}
+
+thread_control_monitor& arena::get_waiting_threads_monitor() {
+    return my_threading_control->get_waiting_threads_monitor();
 }
 
 void arena::enqueue_task(d1::task& t, d1::task_group_context& ctx, thread_data& td) {
@@ -358,6 +460,27 @@ void arena::enqueue_task(d1::task& t, d1::task_group_context& ctx, thread_data&
     advertise_new_work<work_enqueued>();
 }
 
+arena &arena::create(threading_control *control, unsigned num_slots,
+                     unsigned num_reserved_slots, unsigned arena_priority_level,
+                     d1::constraints constraints
+#if __TBB_PREVIEW_PARALLEL_PHASE
+                     , tbb::task_arena::leave_policy lp 
+#endif
+) {
+    __TBB_ASSERT(num_slots > 0, NULL);
+    __TBB_ASSERT(num_reserved_slots <= num_slots, NULL);
+    // Add public market reference for an external thread/task_arena (that adds an internal reference in exchange).
+    arena& a = arena::allocate_arena(control, num_slots, num_reserved_slots, arena_priority_level
+#if __TBB_PREVIEW_PARALLEL_PHASE
+                                     , lp
+#endif
+    );
+    a.my_tc_client = control->create_client(a);
+    // We should not publish arena until all fields are initialized
+    control->publish_client(a.my_tc_client, constraints);
+    return a;
+}
+
 } // namespace r1
 } // namespace detail
 } // namespace tbb
@@ -384,12 +507,12 @@ void assert_arena_priority_valid( tbb::task_arena::priority ) {}
 
 unsigned arena_priority_level( tbb::task_arena::priority a_priority ) {
     assert_arena_priority_valid( a_priority );
-    return market::num_priority_levels - unsigned(int(a_priority) / d1::priority_stride);
+    return d1::num_priority_levels - unsigned(int(a_priority) / d1::priority_stride);
 }
 
 tbb::task_arena::priority arena_priority( unsigned priority_level ) {
     auto priority = tbb::task_arena::priority(
-        (market::num_priority_levels - priority_level) * d1::priority_stride
+        (d1::num_priority_levels - priority_level) * d1::priority_stride
     );
     assert_arena_priority_valid( priority );
     return priority;
@@ -403,6 +526,9 @@ struct task_arena_impl {
     static void wait(d1::task_arena_base&);
     static int max_concurrency(const d1::task_arena_base*);
     static void enqueue(d1::task&, d1::task_group_context*, d1::task_arena_base*);
+    static d1::slot_id execution_slot(const d1::task_arena_base&);
+    static void enter_parallel_phase(d1::task_arena_base*, std::uintptr_t);
+    static void exit_parallel_phase(d1::task_arena_base*, std::uintptr_t);
 };
 
 void __TBB_EXPORTED_FUNC initialize(d1::task_arena_base& ta) {
@@ -433,44 +559,74 @@ void __TBB_EXPORTED_FUNC enqueue(d1::task& t, d1::task_group_context& ctx, d1::t
     task_arena_impl::enqueue(t, &ctx, ta);
 }
 
+d1::slot_id __TBB_EXPORTED_FUNC execution_slot(const d1::task_arena_base& arena) {
+    return task_arena_impl::execution_slot(arena);
+}
+
+void __TBB_EXPORTED_FUNC enter_parallel_phase(d1::task_arena_base* ta, std::uintptr_t flags) {
+    task_arena_impl::enter_parallel_phase(ta, flags);
+}
+
+void __TBB_EXPORTED_FUNC exit_parallel_phase(d1::task_arena_base* ta, std::uintptr_t flags) {
+    task_arena_impl::exit_parallel_phase(ta, flags);
+}
+
 void task_arena_impl::initialize(d1::task_arena_base& ta) {
     // Enforce global market initialization to properly initialize soft limit
     (void)governor::get_thread_data();
+    d1::constraints arena_constraints;
+
+#if __TBB_ARENA_BINDING
+    arena_constraints = d1::constraints{}
+        .set_core_type(ta.core_type())
+        .set_max_threads_per_core(ta.max_threads_per_core())
+        .set_numa_id(ta.my_numa_id);
+#endif /*__TBB_ARENA_BINDING*/
+
     if (ta.my_max_concurrency < 1) {
 #if __TBB_ARENA_BINDING
-
-#if __TBB_PREVIEW_TASK_ARENA_CONSTRAINTS_EXTENSION_PRESENT
-        d1::constraints arena_constraints = d1::constraints{}
-            .set_core_type(ta.core_type())
-            .set_max_threads_per_core(ta.max_threads_per_core())
-            .set_numa_id(ta.my_numa_id);
         ta.my_max_concurrency = (int)default_concurrency(arena_constraints);
-#else /*!__TBB_PREVIEW_TASK_ARENA_CONSTRAINTS_EXTENSION_PRESENT*/
-        ta.my_max_concurrency = (int)default_concurrency(ta.my_numa_id);
-#endif /*!__TBB_PREVIEW_TASK_ARENA_CONSTRAINTS_EXTENSION_PRESENT*/
-
 #else /*!__TBB_ARENA_BINDING*/
         ta.my_max_concurrency = (int)governor::default_num_threads();
 #endif /*!__TBB_ARENA_BINDING*/
     }
 
+#if __TBB_CPUBIND_PRESENT
+    numa_binding_observer* observer = construct_binding_observer(
+        static_cast<d1::task_arena*>(&ta), arena::num_arena_slots(ta.my_max_concurrency, ta.my_num_reserved_slots),
+        ta.my_numa_id, ta.core_type(), ta.max_threads_per_core());
+    if (observer) {
+        // TODO: Consider lazy initialization for internal arena so
+        // the direct calls to observer might be omitted until actual initialization.
+        observer->on_scheduler_entry(true);
+    }
+#endif /*__TBB_CPUBIND_PRESENT*/
+
     __TBB_ASSERT(ta.my_arena.load(std::memory_order_relaxed) == nullptr, "Arena already initialized");
     unsigned priority_level = arena_priority_level(ta.my_priority);
-    arena* a = market::create_arena(ta.my_max_concurrency, ta.my_num_reserved_slots, priority_level, /* stack_size = */ 0);
-    ta.my_arena.store(a, std::memory_order_release);
-    // add an internal market reference; a public reference was added in create_arena
-    market::global_market( /*is_public=*/false);
-#if __TBB_ARENA_BINDING
-    a->my_numa_binding_observer = construct_binding_observer(
-        static_cast<d1::task_arena*>(&ta), a->my_num_slots, ta.my_numa_id, ta.core_type(), ta.max_threads_per_core());
-#endif /*__TBB_ARENA_BINDING*/
+    threading_control* thr_control = threading_control::register_public_reference();
+    arena& a = arena::create(thr_control, unsigned(ta.my_max_concurrency), ta.my_num_reserved_slots,
+                             priority_level, arena_constraints
+#if __TBB_PREVIEW_PARALLEL_PHASE
+                             , ta.get_leave_policy()
+#endif
+    );
+
+    ta.my_arena.store(&a, std::memory_order_release);
+#if __TBB_CPUBIND_PRESENT
+    a.my_numa_binding_observer = observer;
+    if (observer) {
+        observer->on_scheduler_exit(true);
+        observer->observe(true);
+    }
+#endif /*__TBB_CPUBIND_PRESENT*/
 }
 
 void task_arena_impl::terminate(d1::task_arena_base& ta) {
     arena* a = ta.my_arena.load(std::memory_order_relaxed);
     assert_pointer_valid(a);
-    a->my_market->release( /*is_public=*/true, /*blocking_terminate=*/false );
-    a->on_thread_leaving<arena::ref_external>();
+    threading_control::unregister_public_reference(/*blocking_terminate=*/false);
+    a->on_thread_leaving(arena::ref_external);
     ta.my_arena.store(nullptr, std::memory_order_relaxed);
 }
 
@@ -481,15 +637,15 @@ bool task_arena_impl::attach(d1::task_arena_base& ta) {
         arena* a = td->my_arena;
         // There is an active arena to attach to.
         // It's still used by s, so won't be destroyed right away.
-        __TBB_ASSERT(a->my_references > 0, NULL );
+        __TBB_ASSERT(a->my_references > 0, nullptr);
         a->my_references += arena::ref_external;
         ta.my_num_reserved_slots = a->my_num_reserved_slots;
         ta.my_priority = arena_priority(a->my_priority_level);
         ta.my_max_concurrency = ta.my_num_reserved_slots + a->my_max_num_workers;
-        __TBB_ASSERT(arena::num_arena_slots(ta.my_max_concurrency) == a->my_num_slots, NULL);
+        __TBB_ASSERT(arena::num_arena_slots(ta.my_max_concurrency, ta.my_num_reserved_slots) == a->my_num_slots, nullptr);
         ta.my_arena.store(a, std::memory_order_release);
-        // increases market's ref count for task_arena
-        market::global_market( /*is_public=*/true );
+        // increases threading_control's ref count for task_arena
+        threading_control::register_public_reference();
         return true;
     }
     return false;
@@ -511,6 +667,14 @@ void task_arena_impl::enqueue(d1::task& t, d1::task_group_context* c, d1::task_a
      a->enqueue_task(t, *ctx, *td);
 }
 
+d1::slot_id task_arena_impl::execution_slot(const d1::task_arena_base& ta) {
+    thread_data* td = governor::get_thread_data_if_initialized();
+    if (td && (td->is_attached_to(ta.my_arena.load(std::memory_order_relaxed)))) {
+        return td->my_arena_index;
+    }
+    return d1::slot_id(-1);
+}
+
 class nested_arena_context : no_copy {
 public:
     nested_arena_context(thread_data& td, arena& nested_arena, std::size_t slot_index)
@@ -520,19 +684,20 @@ public:
             m_orig_arena = td.my_arena;
             m_orig_slot_index = td.my_arena_index;
             m_orig_last_observer = td.my_last_observer;
+            m_orig_is_thread_registered = td.my_is_registered;
 
             td.detach_task_dispatcher();
             td.attach_arena(nested_arena, slot_index);
+            td.my_is_registered = false;
             if (td.my_inbox.is_idle_state(true))
                 td.my_inbox.set_is_idle(false);
             task_dispatcher& task_disp = td.my_arena_slot->default_task_dispatcher();
-            task_disp.set_stealing_threshold(m_orig_execute_data_ext.task_disp->m_stealing_threshold);
-            td.attach_task_dispatcher(task_disp);
+            td.enter_task_dispatcher(task_disp, m_orig_execute_data_ext.task_disp->m_stealing_threshold);
 
             // If the calling thread occupies the slots out of external thread reserve we need to notify the
             // market that this arena requires one worker less.
             if (td.my_arena_index >= td.my_arena->my_num_reserved_slots) {
-                td.my_arena->my_market->adjust_demand(*td.my_arena, /* delta = */ -1, /* mandatory = */ false);
+                td.my_arena->request_workers(/* mandatory_delta = */ 0, /* workers_delta = */ -1);
             }
 
             td.my_last_observer = nullptr;
@@ -568,14 +733,13 @@ public:
             // Notify the market that this thread releasing a one slot
             // that can be used by a worker thread.
             if (td.my_arena_index >= td.my_arena->my_num_reserved_slots) {
-                td.my_arena->my_market->adjust_demand(*td.my_arena, /* delta = */ 1, /* mandatory = */ false);
+                td.my_arena->request_workers(/* mandatory_delta = */ 0, /* workers_delta = */ 1);
             }
 
-            td.my_task_dispatcher->set_stealing_threshold(0);
-            td.detach_task_dispatcher();
+            td.leave_task_dispatcher();
             td.my_arena_slot->release();
             td.my_arena->my_exit_monitors.notify_one(); // do not relax!
-
+            td.my_is_registered = m_orig_is_thread_registered;
             td.attach_arena(*m_orig_arena, m_orig_slot_index);
             td.attach_task_dispatcher(*m_orig_execute_data_ext.task_disp);
             __TBB_ASSERT(td.my_inbox.is_idle_state(false), nullptr);
@@ -591,6 +755,7 @@ private:
     unsigned            m_orig_slot_index{};
     bool                m_orig_fifo_tasks_allowed{};
     bool                m_orig_critical_task_allowed{};
+    bool                m_orig_is_thread_registered{};
 };
 
 class delegated_task : public d1::task {
@@ -623,7 +788,7 @@ class delegated_task : public d1::task {
     }
     void finalize() {
         m_wait_ctx.release(); // must precede the wakeup
-        m_monitor.notify([this](std::uintptr_t ctx) {
+        m_monitor.notify([this] (std::uintptr_t ctx) {
             return ctx == std::uintptr_t(&m_delegate);
         }); // do not relax, it needs a fence!
         m_completed.store(true, std::memory_order_release);
@@ -712,7 +877,7 @@ void task_arena_impl::wait(d1::task_arena_base& ta) {
     __TBB_ASSERT_EX(td, "Scheduler is not initialized");
     __TBB_ASSERT(td->my_arena != a || td->my_arena_index == 0, "internal_wait is not supported within a worker context" );
     if (a->my_max_num_workers != 0) {
-        while (a->num_workers_active() || a->my_pool_state.load(std::memory_order_acquire) != arena::SNAPSHOT_EMPTY) {
+        while (a->num_workers_active() || !a->is_empty()) {
             yield();
         }
     }
@@ -726,12 +891,12 @@ int task_arena_impl::max_concurrency(const d1::task_arena_base *ta) {
         a = td->my_arena; // the current arena if any
 
     if( a ) { // Get parameters from the arena
-        __TBB_ASSERT( !ta || ta->my_max_concurrency==1, NULL );
-        return a->my_num_reserved_slots + a->my_max_num_workers
-#if __TBB_ENQUEUE_ENFORCED_CONCURRENCY
-            + (a->my_local_concurrency_flag.test() ? 1 : 0)
-#endif
-            ;
+        __TBB_ASSERT( !ta || ta->my_max_concurrency==1, nullptr);
+        int mandatory_worker = 0;
+        if (a->is_arena_workerless() && a->my_num_reserved_slots == 1) {
+            mandatory_worker = a->my_mandatory_concurrency.test() ? 1 : 0;
+        }
+        return a->my_num_reserved_slots + a->my_max_num_workers + mandatory_worker;
     }
 
     if (ta && ta->my_max_concurrency == 1) {
@@ -740,22 +905,33 @@ int task_arena_impl::max_concurrency(const d1::task_arena_base *ta) {
 
 #if __TBB_ARENA_BINDING
     if (ta) {
-#if __TBB_PREVIEW_TASK_ARENA_CONSTRAINTS_EXTENSION_PRESENT
         d1::constraints arena_constraints = d1::constraints{}
             .set_numa_id(ta->my_numa_id)
             .set_core_type(ta->core_type())
             .set_max_threads_per_core(ta->max_threads_per_core());
         return (int)default_concurrency(arena_constraints);
-#else /*!__TBB_PREVIEW_TASK_ARENA_CONSTRAINTS_EXTENSION_PRESENT*/
-        return (int)default_concurrency(ta->my_numa_id);
-#endif /*!__TBB_PREVIEW_TASK_ARENA_CONSTRAINTS_EXTENSION_PRESENT*/
     }
 #endif /*!__TBB_ARENA_BINDING*/
 
-    __TBB_ASSERT(!ta || ta->my_max_concurrency==d1::task_arena_base::automatic, NULL );
+    __TBB_ASSERT(!ta || ta->my_max_concurrency==d1::task_arena_base::automatic, nullptr);
     return int(governor::default_num_threads());
 }
 
+#if __TBB_PREVIEW_PARALLEL_PHASE
+void task_arena_impl::enter_parallel_phase(d1::task_arena_base* ta, std::uintptr_t /*reserved*/) {
+    arena* a = ta ? ta->my_arena.load(std::memory_order_relaxed) : governor::get_thread_data()->my_arena;
+    __TBB_ASSERT(a, nullptr);
+    a->my_thread_leave.register_parallel_phase();
+    a->advertise_new_work<arena::work_enqueued>();
+}
+
+void task_arena_impl::exit_parallel_phase(d1::task_arena_base* ta, std::uintptr_t flags) {
+    arena* a = ta ? ta->my_arena.load(std::memory_order_relaxed) : governor::get_thread_data()->my_arena;
+    __TBB_ASSERT(a, nullptr);
+    a->my_thread_leave.unregister_parallel_phase(/*with_fast_leave=*/static_cast<bool>(flags));
+}
+#endif
+
 void isolate_within_arena(d1::delegate_base& d, std::intptr_t isolation) {
     // TODO: Decide what to do if the scheduler is not initialized. Is there a use case for it?
     thread_data* tls = governor::get_thread_data();
@@ -770,7 +946,7 @@ void isolate_within_arena(d1::delegate_base& d, std::intptr_t isolation) {
         // Isolation within this callable
         d();
     }).on_completion([&] {
-        __TBB_ASSERT(governor::get_thread_data()->my_task_dispatcher == dispatcher, NULL);
+        __TBB_ASSERT(governor::get_thread_data()->my_task_dispatcher == dispatcher, nullptr);
         dispatcher->set_isolation(previous_isolation);
     });
 }
