@@ -1,5 +1,5 @@
 /*
-    Copyright (c) 2005-2022 Intel Corporation
+    Copyright (c) 2005-2025 Intel Corporation
 
     Licensed under the Apache License, Version 2.0 (the "License");
     you may not use this file except in compliance with the License.
@@ -84,7 +84,7 @@ class SharedGroupBodyImpl : utils::NoCopy, utils::NoAfterlife {
 
     void DeleteTaskGroup () {
         delete m_taskGroup;
-        m_taskGroup = NULL;
+        m_taskGroup = nullptr;
     }
 
     void Wait () {
@@ -99,8 +99,10 @@ class SharedGroupBodyImpl : utils::NoCopy, utils::NoAfterlife {
                 utils::ConcurrencyTracker ct;
                 m_taskGroup->wait();
             }
-            if ( utils::ConcurrencyTracker::PeakParallelism() == 1 )
-                WARN( "Warning: No parallel waiting detected in TestParallelWait" );
+            if ( utils::ConcurrencyTracker::PeakParallelism() == 1 ) {
+                const char* msg = "Warning: No parallel waiting detected in TestParallelWait";
+                WARN( msg );
+            }
             m_barrier.wait();
         }
         else
@@ -113,7 +115,7 @@ public:
     SharedGroupBodyImpl ( std::uintptr_t numThreads, std::uintptr_t sharingMode = 0 )
         : m_numThreads(numThreads)
         , m_sharingMode(sharingMode)
-        , m_taskGroup(NULL)
+        , m_taskGroup(nullptr)
         , m_barrier(numThreads)
     {
         CHECK_MESSAGE( m_numThreads > 1, "SharedGroupBody tests require concurrency" );
@@ -253,7 +255,7 @@ void TestThreadSafety() {
     // Test and warm up allocator.
     tests();
 
-    // Ensure that cosumption is stabilized.
+    // Ensure that consumption is stabilized.
     std::size_t initial = utils::GetMemoryUsage();
     for (;;) {
         tests();
@@ -395,7 +397,7 @@ class test_exception : public std::exception
 public:
     test_exception ( const char* descr ) : m_strDescription(descr) {}
 
-    const char* what() const throw() override { return m_strDescription; }
+    const char* what() const noexcept override { return m_strDescription; }
 };
 
 using TestException = test_exception;
@@ -753,7 +755,8 @@ namespace TestIsolationNS {
 //! Test for thread safety for the task_group
 //! \brief \ref error_guessing \ref resource_usage
 TEST_CASE("Memory leaks test is not applicable under ASAN\n" * doctest::skip(true)) {}
-#else
+#elif !EMSCRIPTEN
+//! Emscripten requires preloading of the file used to determine memory usage, hence disabled.
 //! Test for thread safety for the task_group
 //! \brief \ref error_guessing \ref resource_usage
 TEST_CASE("Thread safety test for the task group") {
@@ -777,8 +780,11 @@ TEST_CASE("Thread safety test for the task group") {
 TEST_CASE("Fibonacci test for the task group") {
     for (unsigned p=MinThread; p <= MaxThread; ++p) {
         tbb::global_control limit(tbb::global_control::max_allowed_parallelism, p);
+        tbb::task_arena a(p);
         g_MaxConcurrency = p;
-        RunFibonacciTests<tbb::task_group>();
+        a.execute([] {
+            RunFibonacciTests<tbb::task_group>();
+        });
     }
 }
 
@@ -821,7 +827,7 @@ TEST_CASE("Move semantics test for the task group") {
 //! Test for thread safety for the isolated_task_group
 //! \brief \ref error_guessing
 TEST_CASE("Memory leaks test is not applicable under ASAN\n" * doctest::skip(true)) {}
-#else
+#elif !EMSCRIPTEN
 //! Test for thread safety for the isolated_task_group
 //! \brief \ref error_guessing
 TEST_CASE("Thread safety test for the isolated task group") {
@@ -835,7 +841,10 @@ TEST_CASE("Thread safety test for the isolated task group") {
         }
         tbb::global_control limit(tbb::global_control::max_allowed_parallelism, p);
         g_MaxConcurrency = p;
-        TestThreadSafety<tbb::isolated_task_group>();
+        tbb::task_arena a(p);
+        a.execute([] {
+            TestThreadSafety<tbb::isolated_task_group>();
+        });
     }
 }
 #endif
@@ -846,7 +855,10 @@ TEST_CASE("Fibonacci test for the isolated task group") {
     for (unsigned p=MinThread; p <= MaxThread; ++p) {
         tbb::global_control limit(tbb::global_control::max_allowed_parallelism, p);
         g_MaxConcurrency = p;
-        RunFibonacciTests<tbb::isolated_task_group>();
+        tbb::task_arena a(p);
+        a.execute([] {
+            RunFibonacciTests<tbb::isolated_task_group>();
+        });
     }
 }
 
@@ -856,7 +868,10 @@ TEST_CASE("Cancellation and exception test for the isolated task group") {
     for (unsigned p=MinThread; p <= MaxThread; ++p) {
         tbb::global_control limit(tbb::global_control::max_allowed_parallelism, p);
         g_MaxConcurrency = p;
-        RunCancellationAndExceptionHandlingTests<tbb::isolated_task_group>();
+        tbb::task_arena a(p);
+        a.execute([] {
+            RunCancellationAndExceptionHandlingTests<tbb::isolated_task_group>();
+        });
     }
 }
 
@@ -866,7 +881,10 @@ TEST_CASE("Constant functor test for the isolated task group") {
     for (unsigned p=MinThread; p <= MaxThread; ++p) {
         tbb::global_control limit(tbb::global_control::max_allowed_parallelism, p);
         g_MaxConcurrency = p;
-        TestConstantFunctorRequirement<tbb::isolated_task_group>();
+        tbb::task_arena a(p);
+        a.execute([] {
+            TestConstantFunctorRequirement<tbb::isolated_task_group>();
+        });
     }
 }
 
@@ -876,7 +894,10 @@ TEST_CASE("Move semantics test for the isolated task group") {
     for (unsigned p=MinThread; p <= MaxThread; ++p) {
         tbb::global_control limit(tbb::global_control::max_allowed_parallelism, p);
         g_MaxConcurrency = p;
-        TestMoveSemantics<tbb::isolated_task_group>();
+        tbb::task_arena a(p);
+        a.execute([] {
+            TestMoveSemantics<tbb::isolated_task_group>();
+        });
     }
 }
 
@@ -925,58 +946,60 @@ TEST_CASE("Test for stack overflow avoidance mechanism within arena") {
         return;
     }
 
-    tbb::global_control thread_limit(tbb::global_control::max_allowed_parallelism, 2);
-    tbb::task_group tg1;
-    tbb::task_group tg2;
-    std::atomic<int> tasks_executed{};
+    tbb::task_arena a1(2, 1);
+    a1.execute([] {
+        tbb::task_group tg1;
+        tbb::task_group tg2;
+        std::atomic<int> tasks_executed{};
 
-    // Determine nested task execution limit.
-    int second_thread_executed{};
-    tg1.run_and_wait([&tg1, &tg2, &tasks_executed, &second_thread_executed] {
-        run_deep_stealing(tg1, tg2, 10000, tasks_executed);
-        do {
-            second_thread_executed = tasks_executed;
-            utils::Sleep(10);
-        } while (second_thread_executed < 100 || second_thread_executed != tasks_executed);
-        CHECK(tasks_executed < 10000);
-    });
-    tg2.wait();
-    CHECK(tasks_executed == 10000);
-
-    tasks_executed = 0;
-    tbb::task_arena a(2, 2);
-    tg1.run_and_wait([&a, &tg1, &tg2, &tasks_executed, second_thread_executed] {
-        run_deep_stealing(tg1, tg2, second_thread_executed-1, tasks_executed);
-        while (tasks_executed < second_thread_executed-1) {
-            // Wait until the second thread near the limit.
-            utils::yield();
-        }
-        tg2.run([&a, &tg1, &tasks_executed] {
-            a.execute([&tg1, &tasks_executed] {
-                volatile char consume_stack[1000]{};
-                ++tasks_executed;
-                tg1.wait();
-                utils::suppress_unused_warning(consume_stack);
-            });
-        });
-        while (tasks_executed < second_thread_executed) {
-            // Wait until the second joins the arena.
-            utils::yield();
-        }
-        a.execute([&tg1, &tg2, &tasks_executed] {
+        // Determine nested task execution limit.
+        int second_thread_executed{};
+        tg1.run_and_wait([&tg1, &tg2, &tasks_executed, &second_thread_executed] {
             run_deep_stealing(tg1, tg2, 10000, tasks_executed);
+            do {
+                second_thread_executed = tasks_executed;
+                utils::Sleep(10);
+            } while (second_thread_executed < 100 || second_thread_executed != tasks_executed);
+            CHECK(tasks_executed < 10000);
         });
-        int currently_executed{};
-        do {
-            currently_executed = tasks_executed;
-            utils::Sleep(10);
-        } while (currently_executed != tasks_executed);
-        CHECK(tasks_executed < 10000 + second_thread_executed);
-    });
-    a.execute([&tg2] {
         tg2.wait();
+        CHECK(tasks_executed == 10000);
+
+        tasks_executed = 0;
+        tbb::task_arena a2(2, 2);
+        tg1.run_and_wait([&a2, &tg1, &tg2, &tasks_executed, second_thread_executed] {
+            run_deep_stealing(tg1, tg2, second_thread_executed - 1, tasks_executed);
+            while (tasks_executed < second_thread_executed - 1) {
+                // Wait until the second thread near the limit.
+                utils::yield();
+            }
+            tg2.run([&a2, &tg1, &tasks_executed] {
+                a2.execute([&tg1, &tasks_executed] {
+                    volatile char consume_stack[1000]{};
+                    ++tasks_executed;
+                    tg1.wait();
+                    utils::suppress_unused_warning(consume_stack);
+                });
+            });
+            while (tasks_executed < second_thread_executed) {
+                // Wait until the second joins the arena.
+                utils::yield();
+            }
+            a2.execute([&tg1, &tg2, &tasks_executed] {
+                run_deep_stealing(tg1, tg2, 10000, tasks_executed);
+            });
+            int currently_executed{};
+            do {
+                currently_executed = tasks_executed;
+                utils::Sleep(10);
+            } while (currently_executed != tasks_executed);
+            CHECK(tasks_executed < 10000 + second_thread_executed);
+        });
+        a2.execute([&tg2] {
+            tg2.wait();
+        });
+        CHECK(tasks_executed == 10000 + second_thread_executed);
     });
-    CHECK(tasks_executed == 10000 + second_thread_executed);
 }
 
 //! Test checks that we can submit work to task_group asynchronously with waiting.
@@ -1199,4 +1222,3 @@ TEST_CASE("task_handle cannot be scheduled into other task_group of the same con
 }
 
 #endif // TBB_USE_EXCEPTIONS
-
